#include <iostream>
#include <fstream>
#include <vector>
#include <complex>
#include <cmath>

using namespace std;

const double PI = M_PI;

// FFT implemented by the Cooley-Tukey algorithm
vector<complex<double>> fft(const vector<complex<double>>& x) {
    int N = x.size();
    if (N == 1) {
        return x;
    }

    // Divide
    vector<complex<double>> x0(N / 2), x1(N / 2);
    for (int n = 0; n < N / 2; n++) {
        x0[n] = x[2 * n];
        x1[n] = x[2 * n + 1];
    }

    // Conquer
    vector<complex<double>> X0 = fft(x0);
    vector<complex<double>> X1 = fft(x1);

    // Combine
    vector<complex<double>> X(N);
    double angle = 2 * PI / N;
    for (int m = 0; m < N / 2; m++) {
        complex<double> twiddle_factor = polar(1.0, -angle * m);
        X[m] = X0[m] + twiddle_factor * X1[m];
        X[m + N / 2] = X0[m] - twiddle_factor * X1[m];
    }

    return X;
}

// IFFT
vector<complex<double>> ifft(const vector<complex<double>>& X) {
    int N = X.size();
    vector<complex<double>> FFT_X = fft(X);
    vector<complex<double>> x(N);
    for (int n = 0; n < N; n++) {
        x[n] = FFT_X[(N - n) % N] / static_cast<double>(N);
    }
    return x;
}

// Function to perform overlap-add convolution
vector<complex<double>> overlapAddConvolution(const vector<complex<double>>& x, const vector<complex<double>>& h, int L) {
    // x is the input signal x[n]
    // h is the filter h[n]
    // y is the output signal y[n]

    int N = x.size(); // Length of x[n]
    int M = h.size(); // Length of h[n]
    int S = (N + L - 1) / L;  // Number of sections. S = ceil(N/L)
    int P = L + M - 1; // Number of points in FFT
    vector<complex<double>> y(N + M - 1, 0.0); // y = zeros(N + M - 1)

    // Compute H, the FFT of h
    vector<complex<double>> h_padded(P, 0.0);
    for (int n = 0; n < M; n++) {
        h_padded[n] = h[n];
    }
    vector<complex<double>> H = fft(h_padded);

    for (int s = 0; s < S; ++s) {
        // Extract section of input signal
        // The last section x_{S-1}[n] is zero-padded to length L
        vector<complex<double>> xs(L, 0.0);
        for (int i = 0; i < L; ++i) {
            if (s * L + i < N) {
                xs[i] = x[s * L + i];
            }
        }

        // Convolve the section with the filter using FFT
        // vector<complex<double>> ys = convolveFFT(xs, h);
        vector<complex<double>> xs_padded(P, 0.0);
        for (int n = 0; n < L; n++) {
            xs_padded[n] = xs[n];
        }
        vector<complex<double>> Xs = fft(xs_padded);

        vector<complex<double>> Ys(P);
        for (int m = 0; m < P; m++) {
            Ys[m] = Xs[m] * H[m];
        }

        vector<complex<double>> ys = ifft(Ys);

        // Add the convolved result to the output signal
        for (int i = 0; i < P; ++i) {
            if (s * L + i < y.size()) {
                y[s * L + i] += ys[i];
            }
        }
    }
    return y;
}

vector<complex<double>> readSignal(const string& filename) {
    ifstream file(filename);
    vector<complex<double>> signal;
    double value;
    while (file >> value) {
        signal.push_back(complex<double>(value, 0.0));
    }
    return signal;
}

// Function to write the output signal to a file
void writeSignal(const string& filename, const vector<complex<double>>& signal) {
    ofstream file(filename);
    for (const auto& value : signal) {
        file << value.real() << endl;
    }
}


int main() {
    // Example input signal (complex-valued)
    vector<complex<double>> x = readSignal("./input/input_signal.txt");

    // Example filter (complex-valued)
    vector<complex<double>> h = readSignal("./input/filter_coefficients.txt");

    // Choose the section length L
    int M = h.size();
    int P = 8 * pow(2, ceil(log2(M)));
    int L = P - M + 1;

    // Perform overlap-add convolution
    vector<complex<double>> y = overlapAddConvolution(x, h, L);

    // Write the result to a txt file
    writeSignal("./output/output_signal.txt", y);

    return 0;
}
